@@ -4,6 +4,8 @@
 {-# LANGUAGE ViewPatterns #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE LambdaCase #-}
+{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
+{-# HLINT ignore "Use camelCase" #-}
 module HieDb.Create where
 
 import Prelude hiding (mod)
@@ -15,6 +17,7 @@ import GHC
 
 import Control.Exception
 import Control.Monad
+import Control.Monad.Cont
 import Control.Monad.IO.Class
 import Control.Monad.State.Strict (evalStateT)
 
@@ -37,7 +40,7 @@ import HieDb.Types
 import HieDb.Utils
 
 sCHEMA_VERSION :: Integer
-sCHEMA_VERSION = 8
+sCHEMA_VERSION = 9
 
 dB_VERSION :: Integer
 dB_VERSION = read (show sCHEMA_VERSION ++ "999" ++ show hieVersion)
@@ -59,26 +62,87 @@ checkVersion k db@(getConn -> conn) = do
   else
     throwIO $ IncompatibleSchemaVersion dB_VERSION ver
 
+-- | Commmon implementation for the @withHieDb*@ functions that prepared the
+-- hiedb file, if it didn't already exist. Also prepares any statements that
+-- we'll need to do indexing.
+withHieDbImpl :: FilePath -> (HieDb -> IO a) -> IO a
+withHieDbImpl fp act = withConnection fp $ \connection -> do
+  liftIO $ setupHieDb connection
+
+  -- We manually prepare statements instead of using the machinery in
+  -- sqlite-simple so we can reuse these across `.hie` files more easily.
+  -- This was benchmarked to be faster:
+  --   https://github.com/wz1000/HieDb/pull/86
+  withPreparedHieDbStatements connection $ \statements -> do
+    act $ HieDb connection statements
+
+withPreparedHieDbStatements :: Connection -> (HieDbStatements -> IO a) -> IO a
+withPreparedHieDbStatements connection action = do
+  let prepareInternalTableDeletion = do
+        deletions
+          <- traverse (createStatement connection)
+            [ "DELETE FROM refs  WHERE hieFile = ?"
+            , "DELETE FROM decls WHERE hieFile = ?"
+            , "DELETE FROM defs  WHERE hieFile = ?"
+            , "DELETE FROM typerefs WHERE hieFile = ?"
+            , "DELETE FROM mods  WHERE hieFile = ?"
+            , "DELETE FROM exports WHERE hieFile = ?"
+            ]
+        pure $ \fp -> mapM_ (`runStatementFor_` fp) deletions
+
+  runContT
+    ( HieDbStatements
+        <$> createStatement connection "INSERT INTO mods VALUES (?,?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO refs VALUES (?,?,?,?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO decls VALUES (?,?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO imports VALUES (?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO defs VALUES (?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO exports VALUES (?,?,?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO typerefs VALUES (?,?,?,?,?,?,?)"
+        <*> createStatement connection "INSERT INTO typenames(name,mod,unit) VALUES (?,?,?)"
+        <*> createStatement connection "SELECT id FROM typenames WHERE name = ? AND mod = ? AND unit = ?"
+        <*> prepareInternalTableDeletion
+    )
+    action
+
 {-| Given path to @.hiedb@ file, constructs 'HieDb' and passes it to given function. -}
 withHieDb :: FilePath -> (HieDb -> IO a) -> IO a
-withHieDb fp f = withConnection fp (checkVersion f . HieDb)
+withHieDb fp f = withHieDbImpl fp $ \hiedb -> do
+  checkVersion f hiedb
 
 {-| Given GHC LibDir and path to @.hiedb@ file,
 constructs DynFlags (required for printing info from @.hie@ files)
 and 'HieDb' and passes them to given function.
 -}
 withHieDbAndFlags :: LibDir -> FilePath -> (DynFlags -> HieDb -> IO a) -> IO a
-withHieDbAndFlags libdir fp f = do
+withHieDbAndFlags libdir fp f = withHieDbImpl fp $ \hiedb -> do
   dynFlags <- dynFlagsForPrinting libdir
-  withConnection fp (checkVersion (f dynFlags) . HieDb)
+  checkVersion (f dynFlags) hiedb
 
 {-| Initialize database schema for given 'HieDb'.
 -}
 initConn :: HieDb -> IO ()
-initConn (getConn -> conn) = do
+{-# DEPRECATED initConn "Use setupHieDb instead." #-}
+initConn = setupHieDb . getConn
+
+{-| Initialize database schema for given 'HieDb'.
+-}
+setupHieDb :: Connection -> IO ()
+setupHieDb conn = do
   execute_ conn "PRAGMA busy_timeout = 500;"
   execute_ conn "PRAGMA journal_mode = WAL;"
   execute_ conn "PRAGMA foreign_keys = ON;"
+
+  -- The default setting for synchronous is 'FULL'. `FULL` will issue fsync's
+  -- and wait to ensure that every write is done durably. Switching to `NORMAL`
+  -- moves this to the checkpoints by the WAL (when using a WAL), losing us
+  -- durability. Practically this means that committed transaction may be rolled
+  -- back on system failure. This is fine, as the hiedb file will recover by
+  -- reindexing. See the following.
+  -- - https://sqlite.org/pragma.html#pragma_synchronous
+  -- - https://sqlite.org/wal.html
+  -- - https://github.com/wz1000/HieDb/pull/86 for benchmarks.
+  execute_ conn "PRAGMA synchronous = NORMAL;"
   execute_ conn "PRAGMA defer_foreign_keys = ON;"
 
   execute_ conn "CREATE TABLE IF NOT EXISTS mods \
@@ -117,6 +181,7 @@ initConn (getConn -> conn) = do
                 \, sc   INTEGER NOT NULL \
                 \, el   INTEGER NOT NULL \
                 \, ec   INTEGER NOT NULL \
+                \, is_generated BOOLEAN NOT NULL \
                 \, FOREIGN KEY(hieFile) REFERENCES mods(hieFile) ON UPDATE CASCADE ON DELETE CASCADE DEFERRABLE INITIALLY DEFERRED \
                 \)"
   execute_ conn "CREATE INDEX IF NOT EXISTS refs_mod ON refs(hieFile)"
@@ -185,7 +250,7 @@ Returns an Array mapping 'TypeIndex' to database ID assigned to the
 corresponding record in DB.
 -}
 addArr :: HieDb -> A.Array TypeIndex HieTypeFlat -> IO (A.Array TypeIndex (Maybe Int64))
-addArr (getConn -> conn) arr = do
+addArr hiedb arr = do
   forM arr $ \case
     HTyVarTy n -> addName n
     HTyConApp tc _ -> addName (ifaceTyConName tc)
@@ -198,8 +263,8 @@ addArr (getConn -> conn) arr = do
         let occ = nameOccName n
             mod = moduleName m
             uid = moduleUnit m
-        execute conn "INSERT INTO typenames(name,mod,unit) VALUES (?,?,?)" (occ,mod,uid)
-        fmap fromOnly . listToMaybe <$> query conn "SELECT id FROM typenames WHERE name = ? AND mod = ? AND unit = ?" (occ,mod,uid)
+        runStatementFor_ (insertTypenamesStatement (preparedStatements hiedb)) (occ,mod,uid)
+        fmap fromOnly <$> runStatementFor (queryTypenamesStatement (preparedStatements hiedb)) (occ,mod,uid)
 
 {-| Add references to types from given @.hie@ file to DB. -}
 addTypeRefs
@@ -305,7 +370,7 @@ addRefsFromLoadedInternal
 addRefsFromLoadedInternal
   db@(getConn -> conn) path sourceFile hash skipOptions hf =
     liftIO $ withTransaction conn $ do
-      deleteInternalTables conn path
+      deleteInternalTablesStatement (preparedStatements db) (Only path)
       addRefsFromLoaded_unsafe db path sourceFile hash skipOptions hf
 
 -- | Like 'addRefsFromLoaded' but without:
@@ -314,8 +379,7 @@ addRefsFromLoadedInternal
 --
 --   Mostly useful to index a new database from scratch as fast as possible
 addRefsFromLoaded_unsafe
-  :: MonadIO m
-  => HieDb -- ^ HieDb into which we're adding the file
+  :: HieDb -- ^ HieDb into which we're adding the file
   -> FilePath -- ^ Path to @.hie@ file
   -> SourceFile -- ^ Path to .hs file from which @.hie@ file was created
                 -- Also tells us if this is a real source file?
@@ -323,39 +387,51 @@ addRefsFromLoaded_unsafe
   -> Fingerprint -- ^ The hash of the @.hie@ file
   -> SkipOptions -- ^ Skip indexing certain tables
   -> HieFile -- ^ Data loaded from the @.hie@ file
-  -> m ()
+  -> IO ()
 addRefsFromLoaded_unsafe
- db@(getConn -> conn) path sourceFile hash skipOptions hf = liftIO $ do
+ db path sourceFile hash skipOptions hf = do
 
   let isBoot = "boot" `isSuffixOf` path
       mod    = moduleName smod
       uid    = moduleUnit smod
       smod   = hie_module hf
-      refmap = generateReferencesMap $ getAsts $ hie_asts hf
+      asts   = getAsts $ hie_asts hf
+      refmapAll = generateReferencesMap asts
+      refmapSourceOnly = generateReferencesMap $ fmap (dropNodeInfos GeneratedInfo) asts
+      refmapGeneratedOnly = generateReferencesMap $ fmap (dropNodeInfos SourceInfo) asts
       (srcFile, isReal) = case sourceFile of
         RealFile f -> (Just f, True)
         FakeFile mf -> (mf, False)
       modrow = HieModuleRow path (ModuleInfo mod uid isBoot srcFile isReal hash)
 
-  execute conn "INSERT INTO mods VALUES (?,?,?,?,?,?,?)" modrow
+      -- We want to distinguish between references from source (NodeOrigin is SourceInfo)
+      -- vs. generated by compiler (NodeOrigin is GeneratedInfo).
+      -- Unfortunately generateReferencesMap throws away the info about NodeOrigin,
+      -- so we need to use this to preprocess the ASTs from which the references map is generated.
+      dropNodeInfos :: NodeOrigin -> HieAST a -> HieAST a
+      dropNodeInfos originToDrop (Node (SourcedNodeInfo sniMap) sp children) =
+        let sourceOnlyNodeInfo = SourcedNodeInfo $ M.delete originToDrop sniMap
+        in Node sourceOnlyNodeInfo sp (map (dropNodeInfos originToDrop) children)
 
-  let AstInfo rows decls imports = genAstInfo path smod refmap
+  runStatementFor_ (insertModsStatement (preparedStatements db)) modrow
+
+  let AstInfo refsSrc declsSrc importsSrc = genAstInfo path smod SourceInfo refmapSourceOnly
+      AstInfo refsGen declsGen importsGen = genAstInfo path smod GeneratedInfo refmapGeneratedOnly
 
   unless (skipRefs skipOptions) $
-    executeMany conn "INSERT INTO refs  VALUES (?,?,?,?,?,?,?,?)" rows
+    mapM_ (runStatementFor_ (insertRefsStatement (preparedStatements db))) (refsSrc <> refsGen)
   unless (skipDecls skipOptions) $
-    executeMany conn "INSERT INTO decls VALUES (?,?,?,?,?,?,?)" decls
+    mapM_ (runStatementFor_ (insertDeclsStatement (preparedStatements db))) (declsSrc <> declsGen)
   unless (skipImports skipOptions) $
-    executeMany conn "INSERT INTO imports VALUES (?,?,?,?,?,?)" imports
+    mapM_ (runStatementFor_ (insertImportsStatement (preparedStatements db))) (importsSrc <> importsGen)
 
-  let defs = genDefRow path smod refmap
+  let defs = genDefRow path smod refmapAll
   unless (skipDefs skipOptions) $
-    forM_ defs $ \def ->
-      execute conn "INSERT INTO defs VALUES (?,?,?,?,?,?)" def
+    mapM_ (runStatementFor_ (insertDefsStatement (preparedStatements db))) defs
 
   let exports = generateExports path $ hie_exports hf
   unless (skipExports skipOptions) $
-    executeMany conn "INSERT INTO exports VALUES (?,?,?,?,?,?,?,?)" exports
+    mapM_ (runStatementFor_ (insertExportsStatement (preparedStatements db))) exports
 
   unless (skipTypes skipOptions) $ do
     ixs <- addArr db (hie_types hf)
@@ -385,21 +461,21 @@ removeDependencySrcFiles (getConn -> conn) =
 
 {-| Delete all occurrences of given @.hie@ file from the database -}
 deleteFileFromIndex :: HieDb -> FilePath -> IO ()
-deleteFileFromIndex (getConn -> conn) path = withTransaction conn $ do
-  deleteInternalTables conn path
+deleteFileFromIndex db@(getConn -> conn) path = withTransaction conn $ do
+  deleteInternalTablesStatement (preparedStatements db) (Only path)
 
 {-| Delete all entries associated with modules for which the 'modInfoSrcFile' doesn't exist
 on the disk.
 Doesn't delete it if there is no associated 'modInfoSrcFile'
 -}
 deleteMissingRealFiles :: HieDb -> IO ()
-deleteMissingRealFiles (getConn -> conn) = withTransaction conn $ do
+deleteMissingRealFiles db@(getConn -> conn) = withTransaction conn $ do
   missing_file_keys <- fold_ conn "SELECT hieFile,hs_src FROM mods WHERE hs_src IS NOT NULL AND is_real" [] $
     \acc (path,src) -> do
       exists <- doesFileExist src
       pure $ if exists then acc else path : acc
   forM_ missing_file_keys $ \path -> do
-    deleteInternalTables conn path
+    deleteInternalTablesStatement (preparedStatements db) (Only path)
 
 {-| Garbage collect typenames with no references - it is a good idea to call
 this function after a sequence of database updates (inserts or deletes)
@@ -407,6 +483,9 @@ this function after a sequence of database updates (inserts or deletes)
 garbageCollectTypeNames :: HieDb -> IO Int
 garbageCollectTypeNames (getConn -> conn) = do
   execute_ conn "DELETE FROM typenames WHERE NOT EXISTS ( SELECT 1 FROM typerefs WHERE typerefs.id = typenames.id LIMIT 1 )"
+  -- Tack on a call to optimize that'll vacuum and update any table statistics
+  -- that might have changed. See https://sqlite.org/pragma.html#pragma_optimize.
+  execute_ conn "PRAGMA optimize;"
   changes conn
 
 deleteInternalTables :: Connection -> FilePath -> IO ()
